#!/usr/bin/env python3

import singer
from singer import utils
from singer.catalog import Catalog, CatalogEntry

import dateutil

from tap_datadog_rum.api_client import RUMApiClient
from tap_datadog_rum.schema_mapper import SchemaMapper

REQUIRED_CONFIG_KEYS = ["api_key", "app_key", "start_date"]
MAX_EVENTS_FOR_SCHEMA_INFERENCE = 3000
LOGGER = singer.get_logger()

def generate_all_schemas(config):
    schemas = {}
    for stream_id, stream_config in config['streams'].items():
        config_attribute_mapping = stream_config.get('attribute_mapping') or {}
        schemas[stream_id] = SchemaMapper(config_attribute_mapping).to_schema()
    return schemas

def schemas_to_catalog(schemas):
    streams = []
    for stream_id, schema in schemas.items():
        stream_metadata = [{
            'metadata': {
                'selected': True,
                'replication-method': 'INCREMENTAL',
            },
            'breadcrumb': []
        }]
        streams.append(
            CatalogEntry(
                tap_stream_id=stream_id,
                stream=stream_id,
                schema=schema,
                key_properties=['event_id'],
                metadata=stream_metadata,
                replication_key=None,
                is_view=None,
                database=None,
                table=None,
                row_count=None,
                stream_alias=None,
                replication_method='INCREMENTAL',
            )
        )
    return Catalog(streams)

def discover(config):
    all_schemas = generate_all_schemas(config)
    return schemas_to_catalog(all_schemas)


def sync(client, config, state, catalog):
    """ Sync data from tap source """

    # Loop over selected streams in catalog
    for stream in catalog.get_selected_streams(state):
        LOGGER.info("Syncing stream:" + stream.tap_stream_id)

        singer.write_schema(
            stream_name=stream.tap_stream_id,
            schema=stream.schema.to_dict(),
            key_properties=stream.key_properties,
        )

        stream_config = config['streams'][stream.tap_stream_id]
        query = stream_config['query']
        config_attribute_mapping = stream_config.get('attribute_mapping') or {}
        schema_mapper = SchemaMapper(config_attribute_mapping)
        state_cursor = state.get(stream.tap_stream_id)

        events, next_cursor = client.fetch_events(query, config_attribute_mapping, state_cursor)
        while len(events) > 0:
            mapped_events = schema_mapper.map_events(events)

            singer.write_records(stream.tap_stream_id, mapped_events)

            state[stream.tap_stream_id] = next_cursor
            singer.write_state(state)

            events, next_cursor = client.fetch_events(query, config_attribute_mapping, next_cursor)
    return


@utils.handle_top_exception(LOGGER)
def main():
    # Parse command line arguments
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)
    api_key = args.config['api_key']
    app_key = args.config['app_key']
    start_date = dateutil.parser.parse(args.config['start_date'])
    client = RUMApiClient(api_key, app_key, start_date)

    # If discover flag was passed, run discovery mode and dump output to stdout
    if args.discover:
        catalog = discover(args.config)
        catalog.dump()
        print('')
    # Otherwise run in sync mode
    else:
        if args.catalog:
            catalog = args.catalog
        else:
            catalog = discover(args.config)
        sync(client, args.config, args.state, catalog)


if __name__ == "__main__":
    main()



from datetime import datetime, timezone
from dateutil import parser
from singer.schema import Schema
import re

DEFAULT_ATTRIBUTE_MAPPINGS = {
    'app_env': ('string', 'attributes.tags.env'),
    'app_version': ('string', 'attributes.tags.version'),
    'browser_name': ('string', 'attributes.attributes.browser.name'),
    'browser_useragent': ('string', 'attributes.attributes.session.useragent'),
    'browser_version_major': ('string', 'attributes.attributes.browser.version_major'),
    'browser_version': ('string', 'attributes.attributes.browser.version'),
    'device_brand': ('string', 'attributes.attributes.device.brand'),
    'device_model': ('string', 'attributes.attributes.device.model'),
    'device_name': ('string', 'attributes.attributes.device.name'),
    'device_type': ('string', 'attributes.attributes.device.type'),
    'error_file': ('string', 'attributes.attributes.error.file'),
    'error_fingerprint': ('string', 'attributes.attributes.error.fingerprint'),
    'error_id': ('string', 'attributes.attributes.error.id'),
    'error_message': ('string', 'attributes.attributes.error.message'),
    'error_type': ('string', 'attributes.attributes.error.type'),
    'event_id': ('string', 'id'),
    'event_timestamp': ('datetime', 'attributes.timestamp'),
    'event_type': ('string', 'attributes.attributes.type'),
    'issue_age_ns': ('integer', 'attributes.attributes.issue.age'),
    'issue_first_seen': ('datetime', 'attributes.attributes.issue.first_seen'),
    'issue_id': ('string', 'attributes.attributes.issue.id'),
    'os_name': ('string', 'attributes.attributes.os.name'),
    'os_version_major': ('string', 'attributes.attributes.os.version_major'),
    'os_version': ('string', 'attributes.attributes.os.version'),
    'rum_application_id': ('string', 'attributes.attributes.application.id'),
    'rum_session_id': ('string', 'attributes.attributes.session.id'),
    'service': ('string', 'attributes.service'),
    'user_id': ('string', 'attributes.attributes.usr.id'),
    'view_id': ('string', 'attributes.attributes.view.id'),
    'view_name': ('string', 'attributes.attributes.view.name'),
    'view_referrer': ('string', 'attributes.attributes.view.referrer'),
    'view_url_host': ('string', 'attributes.attributes.view.url_host'),
    'view_url_path_group': ('string', 'attributes.attributes.view.url_path_group'),
    'view_url_path': ('string', 'attributes.attributes.view.url_path'),
    'view_url_scheme': ('string', 'attributes.attributes.view.url_scheme'),
}

def cooerce_string(val):
  if val is None:
    return None

  if type(val) == float:
    str_val = str(val)
    if str_val.endswith('.0'):
      str_val = str_val[:-2]
    return str_val

  if type(val) == datetime:
    return cooerce_datetime(val)

  return str(val)

def cooerce_int(val):
  if type(val) == int:
    return val

  if type(val) == bool:
    return int(val)

  if type(val) == str:
    if val.endswith('.0'):
      val = val[:-2]

    if re.match("^-?\d+$", val):
      return int(val)

  return None

def cooerce_float(val):
  if type(val) == float:
    return val

  if type(val) == int:
    return float(val)

  if type(val) == str:
    try:
      return float(val)
    except ValueError:
      return None

  return None

def cooerce_bool(val):
  if type(val) == bool:
    return val

  if type(val) == int and (val == 0 or val == 1):
    return bool(val)

  if type(val) == str:
    val = val.lower()
    if val in ['true', 't', '1']:
      return True
    if val in ['false', 'f', '0']:
      return False

  return None

def cooerce_datetime(val):
  if type(val) == datetime:
    return val.isoformat()

  if type(val) == str:
    try:
      dt = parser.parse(val)
      return dt.isoformat()
    except:
      return None

  if type(val) == int: # DD emits datetimes as timestamp nanoseconds from epoch
    dt = datetime.fromtimestamp(val / 1000, tz=timezone.utc)
    return dt.isoformat()

  return None

VALUE_COOERCERS = {
  'string': cooerce_string,
  'integer': cooerce_int,
  'boolean': cooerce_bool,
  'number': cooerce_float,
  'datetime': cooerce_datetime
}

def get_nested_attr(dict, key):
    try:
        if "." in key:
            key, rest = key.split(".", 1)
            return get_nested_attr(dict[key], rest)
        else:
            val = dict[key]
            if type(val) == datetime:
                return val.isoformat()
            else:
                return val
    except KeyError:
        return None


# Convert tags from key:value to a dict so they can be accessed like other attributes
def tags_to_dict(tags):
    dict = {}
    for tag in tags:
        key, val = tag.split(':', 1)
        dict[key] = val
    return dict

def format_event(event, attribute_mappings):
    event_dict = event.to_dict()
    event_dict['attributes']['tags'] = tags_to_dict(event_dict['attributes']['tags'])

    formatted = {}
    for attr_name, (attr_type, attr_path) in attribute_mappings.items():
        val = get_nested_attr(event_dict, attr_path)

        cooercer_func = VALUE_COOERCERS.get(attr_type)
        if cooercer_func is None:
          raise("Don't know how to handle attribute type {attr_type}")

        formatted[attr_name] = cooercer_func(val)

    return formatted

class SchemaMapper:
    def __init__(self, attribute_mappings):
      self.attribute_mappings = {**DEFAULT_ATTRIBUTE_MAPPINGS, **attribute_mappings}

    def map_events(self, events):
      return list(map(lambda evt: format_event(evt, self.attribute_mappings), events))

    def to_schema(self):
      properties = {}
      for attr_name, (attr_type, attr_path) in self.attribute_mappings.items():
        if attr_type == 'datetime':
          property = {
            'type': ['null', 'string'],
            'format': 'date-time'
          }
        else:
          property = {'type': sorted([attr_type, 'null'])}
        properties[attr_name] = property
      return Schema.from_dict({'properties': properties, 'type': 'object'})



from time import sleep
import sys

from datadog_api_client import ApiClient, Configuration
from datadog_api_client.exceptions import ApiAttributeError, ApiException
from datadog_api_client.v2.api.rum_api import RUMApi

PAGE_SIZE = 100
MAX_RETRIES = 5

def list_events_with_retries(client, params):
    api_instance = RUMApi(client)
    for _prev_tries in range(0, MAX_RETRIES):
        try:
            return api_instance.list_rum_events(**params)
        except ApiException as err:
            my_err = err
            if err.status < 500 and err.status != 429: # Datadog returns 429 for rate limit
                raise err # Re-raise anything other than a rate limit or 500

            try:
                rate_limit_reset_sec = int(err.headers['x-ratelimit-reset'])
            except (KeyError, ValueError):
                rate_limit_reset_sec = 30

            print(f'Datadog rate limit reached, sleeping for {rate_limit_reset_sec}s', file=sys.stderr)
            sleep(rate_limit_reset_sec)

    raise 'Still hitting Datadog rate limit after all retries exhausted'

class RUMApiClient:
    def __init__(self, api_key, app_key, start_date):
        self.configuration = Configuration()
        self.configuration.api_key["apiKeyAuth"] = api_key
        self.configuration.api_key["appKeyAuth"] = app_key
        self.start_date = start_date

    def fetch_events(self, query, config_attribute_mappings, cursor = None, newest_first = False):
        with ApiClient(self.configuration) as api_client:
            params = {
                'filter_query': query,
                'filter_from': self.start_date,
                'page_limit': PAGE_SIZE,
                'sort': '-timestamp' if newest_first else 'timestamp'
            }
            if cursor:
                params['page_cursor'] = cursor

            response = list_events_with_retries(api_client, params)

            try:
                next_cursor = response.meta.page.after
            except ApiAttributeError:
                next_cursor = None

            return response.data, next_cursor



#!/usr/bin/env python
from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="tap-datadog-rum",
    version="0.7.2",
    description="Singer.io tap for extracting events from the Datadog Real User Monitoring (RUM) API",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="James Shkolnik (js@gusto.com)",
    url="http://github.com/shkolnik/tap-datadog-rum",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_datadog_rum"],
    install_requires=[
        "singer-python",
        "datadog-api-client>=2.2",
        "datadog-api-client<3",
    ],
    entry_points="""
    [console_scripts]
    tap-datadog-rum=tap_datadog_rum:main
    """,
    packages=["tap_datadog_rum"],
    package_data = {},
    include_package_data=False,
)
